
package java.security;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 1998, 2015  All Rights Reserved
 */

/**
 * Checks access to system resources. Supports marking of code
 * as priveleged. Makes context snapshots to allow checking from
 * other contexts.
 *
 * @author		OTI
 * @version		initial
 */

public final class AccessController {
	static {
		// Initialize vm-internal caches
		initializeInternal();
	}

	static final int OBJS_INDEX_ACC = 0;
	static final int OBJS_INDEX_PDS = 1;
	static final int OBJS_ARRAY_SIZE = 3;
	static final int OBJS_INDEX_PERMS_OR_CACHECHECKED = 2;

private static native void initializeInternal();
private static final SecurityPermission createAccessControlContext = new SecurityPermission("createAccessControlContext"); //$NON-NLS-1$

/* [PR CMVC 188787] Enabling -Djava.security.debug option within WAS keeps JVM busy */
static final class DebugRecursionDetection {
	private static ThreadLocal<String> tlDebug = new ThreadLocal<String>();
	static ThreadLocal<String> getTlDebug() {
		return tlDebug;
	}
}

/**
 * Prevents this class from being instantiated.
 */
private AccessController() {
}

/**
 * The object array returned has following format:
 *
 * Pre-JEP140 format: AccessControlContext/ProtectionDomain..., and the length of the object array is NOT divisible by OBJS_ARRAY_SIZE
 * 	First element is an AccessControlContext object which might be null, either from a full permission privileged frame or from current thread if there is no such privileged frames
 * 	ProtectionDomain elements after AccessControlContext object could be in one of following two formats:
 * 	For doPrivileged methods - flag forDoPrivilegedWithCombiner is false:
 * 		the ProtectionDomain element might be null, first ProtectionDomain element is a duplicate of the ProtectionDomain of the caller of doPrivileged
 * 		rest of ProtectionDomain elements are from the callers discovered during stack walking
 * 		the start index of the actual ProtectionDomain element is 2 of the object array returned
 * 	For doPrivilegedWithCombiner methods - flag forDoPrivilegedWithCombiner is true:
 * 		there are only two ProtectionDomain elements, first one is the ProtectionDomain of the caller of doPrivileged
 * 		and the other is the ProtectionDomain of the caller of doPrivilegedWithCombiner
 * 		the fourth element of the object array is NULL for padding to ensure that the length of the object array is NOT divisible by OBJS_ARRAY_SIZE either
 *
 * JEP 140 format: AccessControlContext/ProtectionDomain[]/Permission[]
 * 	The length of the object array is always divisible by OBJS_ARRAY_SIZE.
 * 	Depends on number of limited permission privileged frames, the result are in following format:
 * 		First element is an AccessControlContext object
 * 		Second element could be in one of following two formats:
 * 			For doPrivileged methods - flag forDoPrivilegedWithCombiner is false:
 * 				an array of ProtectionDomain objects in which first ProtectionDomain element is a duplicate of the ProtectionDomain of the caller of doPrivileged
 * 				the start index of the actual ProtectionDomain element is 1 of this ProtectionDomain objects array
 * 			For doPrivilegedWithCombiner methods - flag forDoPrivilegedWithCombiner is true:
 * 				an array of ProtectionDomain objects with only two elements
 * 				first one is the ProtectionDomain of the caller of doPrivileged
 * 				and the other is the ProtectionDomain of the caller of doPrivilegedWithCombiner
 * 		Third element is an array of Limited Permission objects
 * 		Repeating this format:
 * 			AccessControlContext object,
 * 			ProtectionDomain objects array with same format above when flag forDoPrivilegedWithCombiner is false
 * 			 or just the ProtectionDomain of the caller of doPrivileged in case of flag forDoPrivilegedWithCombiner is true
 * 			Permission object array
 * 		 Until a full permission privileged frame or the end of the stack reached.
 *
 * Note: 1. The reason to have Pre-JEP140 and JEP 140 format is to keep similar format and processing logic
 *			when there is no limited doPrivileged method (JEP 140 implementation) involved.
 *			This helped to address performance issue raised by JAZZ 66091: Perf work for LIR 28261 (Limited doPrivileged / JEP 140)
 *		 2. The reason to duplicate the ProtectionDomain object of the caller of doPrivileged is to avoid creating a new object array
 *			without NULL and duplicate ProtectionDomain objects discovered during stack walking while still keeping same order of
 *			those objects.
 *
 * A full permission privileged frame is any frame running one of the following methods:
 *
 * <code><ul>
 * <li>java/security/AccessController.doPrivileged(Ljava/security/PrivilegedAction;)Ljava/lang/Object;</li>
 * <li>java/security/AccessController.doPrivileged(Ljava/security/PrivilegedExceptionAction;)Ljava/lang/Object;</li>
 * <li>java/security/AccessController.doPrivileged(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;</li>
 * <li>java/security/AccessController.doPrivileged(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;</li>
 * </ul></code>
 *
 * A limited permission privileged frame is any frame running one of the following methods:
 *
 * <code><ul>
 * <li>java/security/AccessController.doPrivileged(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;</li>
 * <li>java/security/AccessController.doPrivileged(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;</li>
 * </ul></code>
 *
 * @param depth The stack depth at which to start. Depth 0 is the current frame (the caller of this native).
 * @param forDoPrivilegedWithCombiner The flag to indicate if it is for doPrivilegedWithCombiner method
 *
 * @return an Object[] as description above
 */
private static native Object[] getAccSnapshot(int depth, boolean forDoPrivilegedWithCombiner);

/**
 * This native retrieves the ProtectionDomain object of the non-reflection/MethodHandleInvoke caller as per depth specified.
 * For example, when depth is set to 1 for method doPrivilegedWithCombiner, the ProtectionDomain object of the caller of doPrivilegedWithCombiner is returned
 * Note: it is only for limited doPrivilegedWithCombiner methods for now.
 *
 * @param depth The stack depth at which to start.
 *
 * @return a ProtectionDomain object as per description above
 */
private static native ProtectionDomain getCallerPD(int depth);

/**
 * provide debug info according to debug settings before throwing AccessControlException
 *
 * @param debug		overall debug flag returned from DebugRecursionDetection.getTlDebug()
 * @param perm 		the permission to check
 * @param pDomain	the pDomain to check
 * @param createACCdenied	if true, actual cause of this ACE was SecurityPermission("createAccessControlContext") denied
 * @exception	AccessControlException	always throw an AccessControlException
 */
private static void throwACE(boolean debug, Permission perm, ProtectionDomain pDomain, boolean createACCdenied) {
	if (debug && ((AccessControlContext.debugSetting() & AccessControlContext.DEBUG_ACCESS) != 0)) {
		DebugRecursionDetection.getTlDebug().set("");	//$NON-NLS-1$
		AccessControlContext.debugPrintAccess();
		if (createACCdenied) {
			System.err.println("access denied " + perm + " due to untrusted AccessControlContext since " + createAccessControlContext + " is denied.");	//$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		} else {
			System.err.println("access denied " + perm);	//$NON-NLS-1$
		}
		DebugRecursionDetection.getTlDebug().remove();
	}
	if (debug && ((AccessControlContext.debugSetting() & AccessControlContext.DEBUG_ACCESS_FAILURE) != 0)) {
		DebugRecursionDetection.getTlDebug().set("");	//$NON-NLS-1$
		new Exception("Stack trace").printStackTrace();	//$NON-NLS-1$
		if (createACCdenied) {
			System.err.println("domain that failed " + createAccessControlContext + " check " + pDomain);	//$NON-NLS-1$ //$NON-NLS-2$
		} else {
			System.err.println("domain that failed " + pDomain);	//$NON-NLS-1$
		}
		DebugRecursionDetection.getTlDebug().remove();
	}
	if (createACCdenied) {
		// K002d = Access denied {0} due to untrusted AccessControlContext since {1} is denied
		throw new AccessControlException(com.ibm.oti.util.Msg.getString("K002d", perm, createAccessControlContext), perm);	//$NON-NLS-1$
	} else {
		// K002c = Access denied {0}
		throw new AccessControlException(com.ibm.oti.util.Msg.getString("K002c", perm), perm);	//$NON-NLS-1$
	}
}

/**
 * Helper method to check whether the running program is allowed to access the resource
 * being guarded by the given Permission argument
 * Assuming perm is not null.
 *
 * @param perm the permission to be checked
 * @param acc the AccessControlContext to be checked
 * @param objects the object array returned from native getAccSnapshot
 * @param frame the doPrivileged frame are being checked
 * @param checked the cached check result
 * @param objPDomains the object array containing ProtectionDomain objects
 * @param debug the debug flag
 * @param startPos the start index of the actual ProtectionDomain objects
 * 			1 is JEP140 format, 2 is Pre-JEP140 format
 *
 * @return true if access is granted by a limited permission, otherwise return false
 */
private static boolean checkPermissionHelper(Permission perm, AccessControlContext acc, Object[] objects, int frame, Object[] checked, Object[] objPDomains, boolean debug, int startPos) {
	boolean	limitedPermImplied = false;
	ProtectionDomain[] pDomains = generatePDarray(acc, objPDomains, debug, startPos);
	if (debug && (0 != (AccessControlContext.debugSetting() & AccessControlContext.DEBUG_ACCESS_DOMAIN))) {
		DebugRecursionDetection.getTlDebug().set(""); //$NON-NLS-1$
		AccessControlContext.debugPrintAccess();
		if (null == pDomains || 0 == pDomains.length) {
			System.err.println("domain (context is null)"); //$NON-NLS-1$
		} else {
			for (int i = 0; i < pDomains.length; i++) {
				System.err.println("domain " + i + " " + pDomains[i]); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}
		DebugRecursionDetection.getTlDebug().remove();
	}
	int length = pDomains == null ? 0 : pDomains.length;
	if ((null != acc)
		&& (null != acc.context)
		&& (AccessControlContext.STATE_AUTHORIZED != acc.authorizeState)
		&& (null != System.getSecurityManager())
	) {
		//	startPos: 1 is JEP140 format, 2 is Pre-JEP140 format
		ProtectionDomain callerPD = (ProtectionDomain)objPDomains[startPos - 1];
		if (null != callerPD && !callerPD.implies(createAccessControlContext)) {
			//	new behavior introduced by this fix
			//	an ACE is thrown if there is a untrusted PD but without SecurityPermission createAccessControlContext
			throwACE(debug, perm, callerPD, true);
		}
	}

	if (2 == startPos) {	// Pre-JEP140 format
		if (null != acc && (null != acc.doPrivilegedAcc || null != acc.nextStackAcc || acc.isLimitedContext)) {
			checked = new Object[OBJS_ARRAY_SIZE];	/* checked was null initially when Pre-JEP140 format */
			return AccessControlContext.checkPermissionWithCache(perm, (Object[])pDomains, acc, false, null, null, checked);
		} else {
			for (int i = 0; i < length ; i++) {
				//	invoke PD within acc.context first
				if (null != pDomains && !pDomains[length - i - 1].implies(perm)) {
					throwACE(debug, perm, pDomains[length - i - 1], false);
				}
			}
		}
	} else {
		if (AccessControlContext.checkPermissionWithCache(perm,
				(Object[])pDomains, (AccessControlContext)objects[frame*OBJS_ARRAY_SIZE],
				(null != objects[frame*OBJS_ARRAY_SIZE + OBJS_INDEX_PERMS_OR_CACHECHECKED]),
				(Permission[])objects[frame*OBJS_ARRAY_SIZE + OBJS_INDEX_PERMS_OR_CACHECHECKED],
				null, checked)
		) {
			limitedPermImplied = true;
		}
	}
	return	limitedPermImplied;
}

/**
 * Checks whether the running program is allowed to
 * access the resource being guarded by the given
 * Permission argument.
 *
 * @param		perm					the permission to check
 * @exception	AccessControlException	if access is not allowed.
 * 				NullPointerException if perm is null
 */
public static void checkPermission(Permission perm) throws AccessControlException {
	if (perm == null) {
		throw new NullPointerException();
	}
	boolean		debug = true;
	if (AccessControlContext.debugSetting() != 0) {
		if ((String)DebugRecursionDetection.getTlDebug().get() != null) {
			debug = false;
		}
	}
	if (debug && ((AccessControlContext.debugSetting() & AccessControlContext.DEBUG_ACCESS_STACK) != 0)) {
		DebugRecursionDetection.getTlDebug().set("");	//$NON-NLS-1$
		new Exception("Stack trace").printStackTrace();	//$NON-NLS-1$
		DebugRecursionDetection.getTlDebug().remove();
	}

	Object[] objects = getAccSnapshot(1, false);
	boolean isPreJEP140Format = (0 == objects.length % OBJS_ARRAY_SIZE) ? false : true;

	if (isPreJEP140Format) {
		AccessControlContext acc = (AccessControlContext)objects[0];
		checkPermissionHelper(perm, acc, null, 0, null, objects, debug, 2);	// the actual ProtectionDomain element starts at index 2
	} else {
		int	frameNbr = objects.length/OBJS_ARRAY_SIZE;
		Object[]	checked = new Object[OBJS_ARRAY_SIZE];
		for (int j = 0; j < frameNbr; j++) {
			AccessControlContext acc = (AccessControlContext)objects[j*OBJS_ARRAY_SIZE];
			Object[] objPDomains = (Object[])objects[j*OBJS_ARRAY_SIZE + OBJS_INDEX_PDS];
			if (checkPermissionHelper(perm, acc, objects, j, checked, objPDomains, debug, 1)) {	// the actual ProtectionDomain element starts at index 1
				break;
			}
		}
	}
	if (debug && ((AccessControlContext.debugSetting() & AccessControlContext.DEBUG_ACCESS) != 0)) {
		DebugRecursionDetection.getTlDebug().set("");	//$NON-NLS-1$
		AccessControlContext.debugPrintAccess();
		System.err.println("access allowed " + perm);	//$NON-NLS-1$
		DebugRecursionDetection.getTlDebug().remove();
	}
}

/*
 * Used to keep the context live during doPrivileged().
 *
 * @see 		#doPrivileged(PrivilegedAction, AccessControlContext)
 */
private static void keepalive(AccessControlContext context) {
}
private static void keepalive(Permission... perms) {
}
/**
 * Answers the access controller context of the current thread,
 * including the inherited ones. It basically retrieves all the
 * protection domains from the calling stack and creates an
 * <code>AccessControlContext</code> with them.
 *
 * @return an AccessControlContext which captures the current state
 *
 * @see 		AccessControlContext
 */
public static AccessControlContext getContext() {
	return getContextHelper(false);
}

/**
 * This is a helper method for getContext() & doPrivilegedWithCombiner methods.
 * Answers the access controller context of the current thread including the inherited ones.
 * Refer native getAccSnapshot() for return data format accordingly to the flag forDoPrivilegedWithCombiner.
 *
 * @param forDoPrivilegedWithCombiner The flag to indicate if it is for doPrivilegedWithCombiner method
 * @return an AccessControlContext which captures the current state
 */
@sun.reflect.CallerSensitive
private static AccessControlContext getContextHelper(boolean forDoPrivilegedWithCombiner) {
	Object[] domains = getAccSnapshot(2, forDoPrivilegedWithCombiner);
	boolean isPreJEP140Format = (0 == domains.length % OBJS_ARRAY_SIZE) ? false : true;
	if (isPreJEP140Format) {
		AccessControlContext acc = (AccessControlContext)domains[0];
		// The actual ProtectionDomain element starts at index 2
		ProtectionDomain[] pDomains = generatePDarray(acc, domains, false, 2);
		AccessControlContext accTmp;
		int newAuthorizedState = getNewAuthorizedState(acc, (ProtectionDomain)domains[1]);
		if (null != acc && (null != acc.doPrivilegedAcc || null != acc.nextStackAcc || acc.isLimitedContext)) {
			accTmp = new AccessControlContext(acc, pDomains, newAuthorizedState);
		} else {
			accTmp = new AccessControlContext(pDomains, newAuthorizedState);
		}
		if (null != acc && null != acc.domainCombiner) {
			accTmp.domainCombiner = acc.domainCombiner;
		}
		return	accTmp;
	}
	int	frameNbr = domains.length/OBJS_ARRAY_SIZE;
	AccessControlContext accContext = null;
	AccessControlContext accLower = null;
	for (int j = 0; j < frameNbr; j++) {
		AccessControlContext acc = (AccessControlContext)domains[j*OBJS_ARRAY_SIZE];
		// the actual ProtectionDomain element starts at index 1
		ProtectionDomain[] pDomains = generatePDarray(acc, (Object[])domains[j*OBJS_ARRAY_SIZE + OBJS_INDEX_PDS], false, 1);
		AccessControlContext accTmp;
		int newAuthorizedState = getNewAuthorizedState(acc, (ProtectionDomain)((Object[])domains[j*OBJS_ARRAY_SIZE + OBJS_INDEX_PDS])[0]);
		if (((null != acc) && acc.isLimitedContext) || (1 < frameNbr)) {
			//	there is a limited doPrivilege frame
			accTmp = new AccessControlContext(acc, pDomains, newAuthorizedState);
		} else {
			//	not set doPrivilegedAcc for non limited doPrivilege
			//	all ProtectionDomains already included within pDomains
			accTmp = new AccessControlContext(pDomains, newAuthorizedState);
		}
		if (null != acc && null != acc.domainCombiner) {
			accTmp.domainCombiner = acc.domainCombiner;
		}
		if (null != domains[j*OBJS_ARRAY_SIZE + OBJS_INDEX_PERMS_OR_CACHECHECKED]) {
			//	this is frame with limited permissions
			accTmp.isLimitedContext = true;
			accTmp.limitedPerms = (Permission[])domains[j*OBJS_ARRAY_SIZE + OBJS_INDEX_PERMS_OR_CACHECHECKED];
		}

		if (null == accLower) {
			accContext = accTmp;
		} else {
			accLower.nextStackAcc = accTmp;
		}
		accLower = accTmp;
	}

	return	accContext;
}

/**
 * Helper method to generate ProtectionDomain array for a new AccessControlContext object
 *
 * @param acc the AccessControlContext object to be checked
 * @param domains the incoming ProtectionDomain object array
 * @param debug the debug flag
 * @param startPos the start index of the actual ProtectionDomain object
 *
 * @return a ProtectionDomain object array
 */
private static ProtectionDomain[] generatePDarray(AccessControlContext acc, Object[] domains, boolean debug, int startPos) {
	ProtectionDomain[] pDomains = null;
	if 	(null != acc
		&& null != acc.domainCombiner
		&& ((AccessControlContext.STATE_AUTHORIZED == acc.authorizeState) || null == System.getSecurityManager())
	) {
		if (debug && (0 != (AccessControlContext.debugSetting() & AccessControlContext.DEBUG_ACCESS))) {
			DebugRecursionDetection.getTlDebug().set("");	//$NON-NLS-1$
			AccessControlContext.debugPrintAccess();
			System.err.println("AccessController invoking the Combiner");	//$NON-NLS-1$
			DebugRecursionDetection.getTlDebug().remove();
		}
		pDomains = acc.domainCombiner.combine(toArrayOfProtectionDomains(domains, null, startPos), acc.context);
	} else {
		pDomains = toArrayOfProtectionDomains(domains, acc, startPos);
	}
	if (null != pDomains && 0 == pDomains.length) {
		pDomains = null;
	}
	return pDomains;
}

/**
 * Helper method to determine the newAuthorizedState according to incoming acc & callerPD
 *
 * @param	acc the AccessControlContext object to be checked
 * @param	callerPD the caller's ProtectionDomain object
 *
 * @return	 AccessControlContext.STATE_AUTHORIZED or STATE_NOT_AUTHORIZED (can't be STATE_UNKNOWN)
 */
private static int getNewAuthorizedState(AccessControlContext acc, ProtectionDomain callerPD) {
	int newAuthorizedState;
	if ((null != acc) && (null != System.getSecurityManager())) {
		newAuthorizedState = acc.authorizeState;
		if (AccessControlContext.STATE_UNKNOWN == newAuthorizedState) {
			//	only change AccessControlContext.authorizeState when it is unknown initially
			if (null == callerPD || callerPD.implies(createAccessControlContext)) {
				newAuthorizedState = AccessControlContext.STATE_AUTHORIZED;
			} else {
				newAuthorizedState = AccessControlContext.STATE_NOT_AUTHORIZED;
			}
		}
	} else {
		newAuthorizedState = AccessControlContext.STATE_AUTHORIZED;
	}
	return	newAuthorizedState;
 }

/**
 * Helper method to combine the ProtectionDomain objects
 *
 * @param domains the incoming ProtectionDomain object array
 * @param acc the AccessControlContext whose context is an ProtectionDomain object array
 * @param startPos the start index of the domains
 *
 * @return an ProtectionDomain object array
 *
 */
private static ProtectionDomain[] toArrayOfProtectionDomains(Object[] domains, AccessControlContext acc, int startPos) {
	int len = 0;
	int extra = 0;
	int size = (null == domains) ? 0 : domains.length;
	if (null != acc && null != acc.context) {
		extra = acc.context.length;
	}
	ProtectionDomain[] answer = new ProtectionDomain[size + extra];
	for (int i = startPos; i < size; i++) {
		boolean found = false;
		answer[len] = (ProtectionDomain)domains[i];
		if (null == answer[len]) {
			break;
		}
		if (null != acc && null != acc.context) {
			for (int j = 0; j < acc.context.length; j++) {
				if (answer[len] == acc.context[j]) {
					found = true;
					break;
				}
			}
		}
		if (!found) {
			len++;
		}
	}
	if (0 == len && null != acc) {
		return acc.context;
	} else {
		if (0 == (len + extra)) {
			return null;
		}
	}
	if (len < size) {
		ProtectionDomain[] copy = new ProtectionDomain[len + extra];
		System.arraycopy(answer, 0, copy, 0, len);
		answer = copy;
	}
	if (null != acc && null != acc.context) {
		System.arraycopy(acc.context, 0, answer, len, acc.context.length);
	}
	return answer;
}

/**
 * Performs the privileged action specified by <code>action</code>.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted. In otherwords, the check stops here.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * @param <T> the type of value returned by PrivilegedAction.run
 *
 * @param action The PrivilegedAction to performed
 *
 * @return the result of the PrivilegedAction
 *
 * @exception	NullPointerException if action is null
 *
 * @see 		#doPrivileged(PrivilegedAction)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivileged(PrivilegedAction<T> action) {
	return action.run();
}

/**
 * Performs the privileged action specified by <code>action</code>.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted iff it is granted by the AccessControlContext
 * <code>context</code>. In otherwords, no more checking of the current stack
 * is performed. Instead, the passed in context is checked.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * @param <T> the type of value returned by PrivilegedAction.run
 *
 * @param action The PrivilegedAction to performed
 * @param context The AccessControlContext to check
 *
 * @return the result of the PrivilegedAction
 *
 * @exception	NullPointerException if action is null
 *
 * @see 		#doPrivileged(PrivilegedAction)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivileged(PrivilegedAction<T> action, AccessControlContext context) {
	T result = action.run();
	keepalive(context);
	return result;
}

/**
 * Performs the privileged action specified by <code>action</code>.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted. In otherwords, the check stops here.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * However, checked exceptions will be caught an re-thrown as PrivilegedActionExceptions
 * @param <T> the type of value returned by PrivilegedExceptionAction.run
 *
 * @param action The PrivilegedExceptionAction to performed
 *
 * @return the result of the PrivilegedExceptionAction
 *
 * @throws PrivilegedActionException when a checked exception occurs when performing the action
 * 			NullPointerException if action is null
 *
 * @see 		#doPrivileged(PrivilegedAction)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivileged(PrivilegedExceptionAction<T> action)
	throws PrivilegedActionException
{
	try {
		return action.run();
	} catch (RuntimeException ex) {
		throw ex;
	} catch (Exception ex) {
		throw new PrivilegedActionException(ex);
	}
}

/**
 * Performs the privileged action specified by <code>action</code>.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted iff it is granted by the AccessControlContext
 * <code>context</code>. In otherwords, no more checking of the current stack
 * is performed. Instead, the passed in context is checked.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * However, checked exceptions will be caught an re-thrown as PrivilegedActionExceptions
 * @param <T> the type of value returned by PrivilegedExceptionAction.run
 *
 * @param action The PrivilegedExceptionAction to performed
 * @param context The AccessControlContext to check
 *
 * @return the result of the PrivilegedExceptionAction
 *
 * @throws PrivilegedActionException when a checked exception occurs when performing the action
 * 			NullPointerException if action is null
 *
 * @see 		#doPrivileged(PrivilegedAction)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivileged (PrivilegedExceptionAction<T> action, AccessControlContext context)
	throws PrivilegedActionException
{
	try {
		T result = action.run();
		keepalive(context);
		return result;
	} catch (RuntimeException ex) {
		throw ex;
	} catch (Exception ex) {
		throw new PrivilegedActionException(ex);
	}
}

/**
 * Performs the privileged action specified by <code>action</code>, retaining
 * any current DomainCombiner.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted. In otherwords, the check stops here.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * @param <T> the type of value returned by PrivilegedAction.run
 *
 * @param action The PrivilegedAction to performed
 *
 * @return the result of the PrivilegedAction
 *
 * @see 		#doPrivileged(PrivilegedAction)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivilegedWithCombiner(PrivilegedAction<T> action) {
	return doPrivileged(action, getContextHelper(true));
}

/**
 * Performs the privileged action specified by <code>action</code>, retaining
 * any current DomainCombiner.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted. In otherwords, the check stops here.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * However, checked exceptions will be caught an re-thrown as PrivilegedActionExceptions
 * @param <T> the type of value returned by PrivilegedExceptionAction.run
 *
 * @param action The PrivilegedExceptionAction to performed
 *
 * @return the result of the PrivilegedExceptionAction
 *
 * @throws PrivilegedActionException when a checked exception occurs when performing the action
 *
 * @see 		#doPrivileged(PrivilegedAction)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivilegedWithCombiner(PrivilegedExceptionAction<T> action)
	throws PrivilegedActionException
{
	return doPrivileged(action, getContextHelper(true));
}

/**
 * Helper method to check if any permission is null
 *
 * @param perms The Permission arguments to limit the scope of the caller's privileges.
 *
 * @exception	NullPointerException if any perm is null
 *
 */
private static void checkPermsNPE(Permission... perms) {
	for (int i = 0; i < perms.length; i++) {
		if (null == perms[i]) {
			throw new NullPointerException();
		}
	}
}

/**
 * Performs the privileged action specified by <code>action</code>.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted iff it is granted by the AccessControlContext
 * <code>context</code> and also granted by one of the permissions arguments.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * @param <T> the type of value returned by PrivilegedAction.run
 *
 * @param action The PrivilegedAction to performed
 * @param context The AccessControlContext to check
 * @param perms The Permission arguments to limit the scope of the caller's privileges.
 *
 * @return the result of the PrivilegedAction
 * @since 1.8
 *
 * @exception	NullPointerException if action is null
 *
 * @see 		#doPrivileged(PrivilegedAction)
 * @see 		#doPrivileged(PrivilegedAction, AccessControlContext)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivileged(PrivilegedAction<T> action,
        AccessControlContext context, Permission... perms)
{
	checkPermsNPE(perms);
	T result = action.run();
	keepalive(context);
	keepalive(perms);
	return result;
}

/**
 * Performs the privileged action specified by <code>action</code>, retaining
 * any current DomainCombiner.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted iff it is granted by one of the permissions arguments.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * @param <T> the type of value returned by PrivilegedAction.run
 *
 * @param action The PrivilegedAction to performed
 * @param context The AccessControlContext to check
 * @param perms The Permission arguments to limit the scope of the caller's privileges.
 *
 * @return the result of the PrivilegedAction
 * @since 1.8
 *
 * @see 		#doPrivileged(PrivilegedAction)
 * @see 		#doPrivileged(PrivilegedAction, AccessControlContext)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivilegedWithCombiner(PrivilegedAction<T> action,
        AccessControlContext context, Permission... perms)
{
	checkPermsNPE(perms);
	ProtectionDomain domain = getCallerPD(1);
	return doPrivileged(action, new AccessControlContext(context, new ProtectionDomain[] {domain}, getNewAuthorizedState(context, domain)), perms);
}

/**
 * Performs the privileged action specified by <code>action</code>.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted iff it is granted by the AccessControlContext
 * <code>context</code> and also granted by one of the permissions arguments.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * However, checked exceptions will be caught an re-thrown as PrivilegedActionExceptions
 * @param <T> the type of value returned by PrivilegedExceptionAction.run
 *
 * @param action The PrivilegedExceptionAction to performed
 * @param context The AccessControlContext to check
 * @param perms The Permission arguments to limit the scope of the caller's privileges.
 *
 * @return the result of the PrivilegedExceptionAction
 * @since 1.8
 *
 * @throws PrivilegedActionException when a checked exception occurs when performing the action
 * 			NullPointerException if action is null
 *
 * @see 		#doPrivileged(PrivilegedAction)
 * @see 		#doPrivileged(PrivilegedAction, AccessControlContext)
 */
@sun.reflect.CallerSensitive
public static <T> T doPrivileged(PrivilegedExceptionAction<T> action,
        AccessControlContext context, Permission... perms)
    throws PrivilegedActionException
{
	try {
		checkPermsNPE(perms);
		T result = action.run();
		keepalive(context);
		keepalive(perms);
		return result;
	} catch (RuntimeException ex) {
		throw ex;
	} catch (Exception ex) {
		throw new PrivilegedActionException(ex);
	}
}

/**
 * Performs the privileged action specified by <code>action</code>, retaining
 * any current DomainCombiner.
 * <p>
 * When permission checks are made, if the permission has been granted by all
 * frames below and including the one representing the call to this method,
 * then the permission is granted and also granted by one of the permissions arguments.
 *
 * Any unchecked exception generated by this method will propagate up the chain.
 * However, checked exceptions will be caught an re-thrown as PrivilegedActionExceptions
 * @param <T> the type of value returned by PrivilegedExceptionAction.run
 *
 * @param action The PrivilegedExceptionAction to performed
 * @param context The AccessControlContext to check
 * @param perms The Permission arguments to limit the scope of the caller's privileges.
 *
 * @return the result of the PrivilegedExceptionAction
 * @since 1.8
 *
 * @throws PrivilegedActionException when a checked exception occurs when performing the action
 *
 * @see 		#doPrivileged(PrivilegedAction)
 * @see 		#doPrivileged(PrivilegedAction, AccessControlContext)
 */
public static <T> T doPrivilegedWithCombiner(PrivilegedExceptionAction<T> action,
        AccessControlContext context, Permission... perms)
  throws PrivilegedActionException
{
	checkPermsNPE(perms);
	ProtectionDomain domain = getCallerPD(1);
	return doPrivileged(action, new AccessControlContext(context, new ProtectionDomain[] {domain}, getNewAuthorizedState(context, domain)), perms);
}

}
